(* C front-end
Copyright (c) 2001 David Monniaux

See the BSD.LICENSE for licensing issues. *)

(*** Lexer interface *)

(* This module is NOT reentrant *)

(**** The extent of the syntactic element in the file *)

(* A locus is a location in the source text. *)
type locus = {
  locus_file_name   : string;
  locus_file_line   : int;
  locus_file_column : int;
  locus_file_char   : int }

(* An extent is a pair of loci, giving the beginning and the end of
   the syntactic element. *)
type extent = locus*locus

let extent_file_name extent = (Filename.basename (fst extent).locus_file_name)
let extent_file_line extent = (fst extent).locus_file_line

(* (fromto t1 t2) is the combined extent of t1 and t2;
     t1 and t2 are meant to be the leftmost and rightmost subexpressions of
     a parsed expression *)

let fromto ((_, (ebeg, _)) : ('a * extent))
    ((_, (_, eend)) : ('b * extent)) =
    ((ebeg, eend) : extent)

(* How to find locations in source *)
let current_line_num       = ref 1
let current_line_start_pos = ref 0
let current_file_name      = ref "*UNKNOWN*"

let beginning_of_file ?(line_offset=1) filename =
  current_line_num       := line_offset;
  current_line_start_pos := 0;
  current_file_name      := filename;;

let next_line lexbuf =
  incr current_line_num;
  current_line_start_pos := Lexing.lexeme_start lexbuf;;

let find_locus_in_current_line n =
  { locus_file_name   = !current_file_name;
    locus_file_line   = !current_line_num;
    locus_file_column = n - !current_line_start_pos;
    locus_file_char   = n }

let extent lexbuf =
  (( (find_locus_in_current_line (Lexing.lexeme_start lexbuf)),
     (find_locus_in_current_line (Lexing.lexeme_end lexbuf)) ) : extent)

(*** Error handling *)
exception Error
exception Lexical_error of locus
exception Internal_error

let lexing_error lexbuf =
  raise (Lexical_error(find_locus_in_current_line
			 (Lexing.lexeme_start lexbuf)));;

let print_warning (ebeg, eend) pp v =
  (* Error message format defined in GNU Coding Standards [2001/10/19] *)
  Printf.fprintf stderr "%s:%d:%d: warning: %a\n"
    ebeg.locus_file_name
    ebeg.locus_file_line
    ebeg.locus_file_column
    pp v;;

let print_error (ebeg, eend) pp v =
  (* Error message format defined in GNU Coding Standards [2001/10/19] *)
  Printf.fprintf stderr "%s:%d:%d: %a\n"
    ebeg.locus_file_name
    ebeg.locus_file_line
    ebeg.locus_file_column
   pp v;;
 
let syntax_error where =
  (* Error message format defined in GNU Coding Standards [2001/10/19] *)
  Printf.fprintf stderr "%s:%d: syntax error in %s\n"
    !current_file_name
    !current_line_num
    where;;
